/***********************************************************************************
FiltroDifusaoComplexa.java - realce de imagens por difuso complexa linear.

Direitos Autorais Reservados (c) 2006 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

**********************************************************************************

Parte integrante do Comunicado Tcnico 075/06.

Descrio:

  Implementa o filtro de realce de imagens e eliminao de rudos usando um filtro
  linear de difuso complexa.

  A implementao do filtro tem por referncia o artigo:

  GILBOA, G.; SOCHE, N.; ZEEVI, Y. Y. Image enhancement and denoising by complex 
    diffusion processes. IEEE Transactions on Pattern Analysis and Machine 
    Intelligence. 26(8):1020-1036, 2004.

  Este programa depende das classes Complex.java e MatrizComplexa.java, encontrados
  neste diretrio.

Parmetros de entrada:

  1. Nome do arquivo imagem.
  2. Nmero de interaes. Equivale ao parmetro 'tempo' na equao (3) do 
     Comunicado Tcnico.

Sada:

  Painel com a imagem original e as imagens com os fatores imaginrio e real
  da imagem complexa.

Desenvolvido por:

    Jos Iguelmar Miranda & Joo Camargo Neto.

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

***********************************************************************************/

// pacotes genricos
import java.io.File;
import java.io.IOException;

// pacotes do AWT
import java.awt.image.WritableRaster;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.GridLayout;

// pacote Swing para interface grfica
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.ImageIcon;
import javax.swing.JScrollPane;

// leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;
	
public class FiltroDifusaoComplexa extends JFrame {

	// ngulo de fase do coeficiente de difuso
	double theta = Math.PI/1000;			

  public static void main(String args[]) {
		int interacao = 0;

		// Certifica a passagem de parmetros
    if (args.length != 2) {
			String msga = "Uso: java -cp . FiltroDifusaoComplexa ";
			String msgb = "<imagem> <interacoes>";
			System.out.println(msga + msgb);
      System.exit(0);
    }

		// Mostra JFrame decorado pelo Swing
		JFrame.setDefaultLookAndFeelDecorated(true);

		long eq_time = System.currentTimeMillis();

		try {
        interacao = Integer.parseInt(args[1]);
    } catch (NumberFormatException e) {
				String st = "Valor do parametro 'interacoes' invalido";
        System.out.println(st);
        System.exit(0);
    }

		if (interacao <= 0)	{
			System.out.println("parametro 'interacao' deve ser > 0");
		}

		System.out.println("\nFiltro de difusao complexa - Parametros:\n");
		String st = "\t#interacoes: ";
	  System.out.println(st + interacao);

		FiltroDifusaoComplexa c = 
			new FiltroDifusaoComplexa(args[0], interacao); 

    eq_time = System.currentTimeMillis() - eq_time;
		String msg = "FiltroComplexo: tempo de execucao ";
    System.out.println(msg + eq_time + " milisseg.");

		// Encerra a aplicao clicando no "close"
		c.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		c.setVisible(true);
	}
  
  /*  Observao:
   *
   *    No CT, o construtor do filtro est definido como  
   *    public FiltroDifusaoComplexa(String aFile, int interacao, int metodo, 
   *                                 int limiar)
   *    Como resolvemos separar os tipos de filtros (linear e no linear),
   *    o construtor teve que ser alterado. Como a diagramao final do artigo
   *    j estava disponibilizado para o pblico, achamos mais prtico fazer esta 
   *    ementa. Isso em nada altera o CT sobre difuso linear.
   *    O leitor encontra um CT dedicado ao caso no linear em separado.
  */

  public FiltroDifusaoComplexa(String aFile, int interacao) {
		
		// Define objeto BufferedImage para encapsular a imagem
    BufferedImage src = null, dest = null, dest1 = null;
		int w, h, tipo, limiarR, pixels[][];
		int[][] bordas;
		Complex limiarJ;
		JLabel img1, img2, img3;

		// Certifica que a imagem existe
    File file = new File(aFile);
    try {
        src = ImageIO.read(file);
    } catch(Exception e) {
        System.out.println("Imagem '" + aFile + "' nao existe.");
        System.exit(0);
    }

    // Atribui nome ao frame
    setTitle("Difuso Complexa: " + file.getName());

		w = src.getWidth();
		h = src.getHeight();
		
		// Mtodo linear.
    // Retorna a parte imaginria.
		limiarJ = new Complex(1, theta);

    // limiar = K = 1 (Gilboa) => retorna a parte real.
		limiarR = 1;  											
		pixels = lePixels(src);

		// Mostra fator imaginrio da imagem complexa.
		bordas = difusaoComplexa(pixels, interacao, limiarJ);
		dest = criaImagem(bordas);

		// Mostra fator real da imagem complexa.
		bordas = difusaoComplexa(pixels, interacao, limiarR);
		dest1 = criaImagem(bordas);

		// Cria gridLayout de 1 x 3
	  getContentPane().setLayout(new GridLayout(1, 3));
		img1 = new JLabel(new ImageIcon(src));
		img2 = new JLabel(new ImageIcon(dest));
		img3 = new JLabel(new ImageIcon(dest1));
		getContentPane().add(new JScrollPane(img1));
		getContentPane().add(new JScrollPane(img2));
		getContentPane().add(new JScrollPane(img3));
		setSize(3*w, h);
	}

	private int[][] lePixels(BufferedImage src) {
		int w, h, pixels[][] = null;

		w = src.getWidth();
		h = src.getHeight();
		pixels = new int[w][h];
    Raster srcR = src.getRaster(); 

		for (int i = 0; i < h; i++)
			for (int j = 0; j < w; j++) {
				pixels[j][i] = srcR.getSample(j, i, 0);
			}

		return pixels;
	}

  // Difuso complexa: retorna a parte imaginria
	private int[][] difusaoComplexa(int[][] in, int interacoes, Complex limiar){
		int h, w, bfi[][];
		double dt, buferMatriz[][];
		
	  w = in.length;
		h = in[0].length;  
		bfi  = new int[w][h];
	  dt = 0.2;
		
		//Criao dos objetos MatrizComplexa a partir da imagen
		MatrizComplexa m1ComplexaJ = new MatrizComplexa(in);
		MatrizComplexa m1ComplexaN = new MatrizComplexa(in);
		MatrizComplexa m1ComplexaS = new MatrizComplexa(in);
		MatrizComplexa m1ComplexaL = new MatrizComplexa(in);
		MatrizComplexa m1ComplexaO = new MatrizComplexa(in);

		System.out.print("\nDifusao complexa: parte imaginaria\n");
		System.out.println("Processando...");

		for (int i = 0; i < interacoes; i++){		
			m1ComplexaN.gradiente(1);
			m1ComplexaS.gradiente(2);
			m1ComplexaL.gradiente(3);
			m1ComplexaO.gradiente(4);

			System.out.print(".");

			m1ComplexaN.multiplica(limiar);
			m1ComplexaS.multiplica(limiar);
			m1ComplexaL.multiplica(limiar);
			m1ComplexaO.multiplica(limiar);
			
			// J = J + dt*(C.*In + C.*Is + C.*Ie + C.*Iw);
			m1ComplexaN.soma(m1ComplexaS);
			m1ComplexaO.soma(m1ComplexaL);
			m1ComplexaN.soma(m1ComplexaO);
			m1ComplexaN.multiplica(dt);
			m1ComplexaJ.soma(m1ComplexaN);
			
			m1ComplexaN.copia(m1ComplexaJ);
			m1ComplexaS.copia(m1ComplexaJ);
			m1ComplexaL.copia(m1ComplexaJ);
			m1ComplexaO.copia(m1ComplexaJ);			
		}

		buferMatriz = m1ComplexaJ.retornaParteImaginaria();
		for(int lin = 0; lin < h; lin++){
			for(int col = 0; col < w; col++){
				bfi[col][lin] = (int) (128.0 + buferMatriz[col][lin]/theta);      		 
			} 
		}

		// Libera rea de armazenamento.
		m1ComplexaN = null;
		m1ComplexaS = null;
		m1ComplexaL = null;
		m1ComplexaO = null;
		m1ComplexaJ = null;
		
		System.out.println("\n");
		return bfi;
	}

  // Difuso complexa: retorna a parte real
	private int[][] difusaoComplexa(int[][] in, int interacoes, int limiar){
		int h, w, bfi[][];
		double dt;
		double[][] buferMatriz;
		
	  w = in.length;
		h = in[0].length;  
		bfi  = new int[w][h];
	  dt = 0.2;
		
		//Criao dos objetos MatrizComplexa a partir da imagen
		MatrizComplexa m1ComplexaJ = new MatrizComplexa(in);
		MatrizComplexa m1ComplexaN = new MatrizComplexa(in);
		MatrizComplexa m1ComplexaS = new MatrizComplexa(in);
		MatrizComplexa m1ComplexaL = new MatrizComplexa(in);
		MatrizComplexa m1ComplexaO = new MatrizComplexa(in);
			
		System.out.print("\nDifusao complexa: parte real\n");
		System.out.println("Processando...");

		for (int i = 0; i < interacoes; i++){		
			m1ComplexaN.gradiente(1);
			m1ComplexaS.gradiente(2);
			m1ComplexaL.gradiente(3);
			m1ComplexaO.gradiente(4);

			System.out.print(".");
					
			// J = J + dt*(C.*In + C.*Is + C.*Ie + C.*Iw);
			// C = limiar
			m1ComplexaN.multiplica((double) limiar);
			m1ComplexaS.multiplica((double) limiar);
			m1ComplexaL.multiplica((double) limiar);
			m1ComplexaO.multiplica((double) limiar);
			
			m1ComplexaN.soma(m1ComplexaS);
			m1ComplexaO.soma(m1ComplexaL);
			m1ComplexaN.soma(m1ComplexaO);
			m1ComplexaN.multiplica(dt);
			m1ComplexaJ.soma(m1ComplexaN);
			
			m1ComplexaN.copia(m1ComplexaJ);
			m1ComplexaS.copia(m1ComplexaJ);
			m1ComplexaL.copia(m1ComplexaJ);
			m1ComplexaO.copia(m1ComplexaJ);			
		}

		buferMatriz = m1ComplexaJ.retornaParteReal();
		for(int lin = 0; lin < h; lin++){
			for(int col = 0; col < w; col++){
				bfi[col][lin] = (int) Math.abs(buferMatriz[col][lin]);      		 
			} 
		}

		// Libera rea de armazenamento.
		m1ComplexaN = null;
		m1ComplexaS = null;
		m1ComplexaL = null;
		m1ComplexaO = null;
		m1ComplexaJ = null;

		System.out.println("\n");
		return bfi;
	}

	private BufferedImage criaImagem(int[][] in) {
		int w, h, tipo, pixels[], ind = 0;
		BufferedImage dest = null;

	  w = in.length;
		h = in[0].length;
		pixels = new int[w*h];
    tipo = BufferedImage.TYPE_BYTE_GRAY;
    dest = new BufferedImage(w, h, tipo);
		WritableRaster destWR = dest.getRaster();

		for(int lin = 0; lin < h; lin++)
			for(int col = 0; col < w; col++)	
				pixels[ind++] = in[col][lin];

		destWR.setPixels(0, 0, w, h, pixels);
		return dest;		
	}

}